#!/usr/bin/env python3
'''
k3testharness is a framework for hosting test.
'''

import argparse
import importlib.util
import sys
import os
import inspect
import logging
import json
import datetime

import k3testharness.api.exceptions
from k3testharness import api

TEST_OK = 0
TEST_FAILED = 1
TESTING_FAILED = 2

logger = logging.getLogger(__name__)

__version__ = '2.0'
__author__ = 'Joachim Kestner <joachim.kestner@khoch3.de>'



# #########################
# NEW LOGGING

def setLogArgs(parser):
    parser.add_argument("-v", "--verbose", action="store_true", help="Enable info logging to console")
    parser.add_argument("-vv", "--extra_verbose", action="store_true", help="Enable debug logging to console")
    #parser.add_argument("--log_file", default=datetime.datetime.now().strftime("%Y%m%dT%H%M%S")+"_"+logFileSuffix, help="Log file with debug logging. Default: <Datetime>_"+logFileSuffix)

def evalLogArgsAInitLogging(args, outputDir):
    fileDebugLogH = logging.FileHandler(os.path.join(outputDir, "debug.log"))
    fileInfoLogH = logging.FileHandler(os.path.join(outputDir, "info.log"))
    fileInfoLogH.setLevel(logging.INFO)
    streamLogH = logging.StreamHandler(sys.stdout)
    if args.verbose:
        streamLogH.setLevel(logging.INFO)
    elif args.extra_verbose:
        streamLogH.setLevel(logging.DEBUG)
    else:
        streamLogH.setLevel(logging.WARN)
    
    fmtStrDebug = "%(asctime)s %(levelname)7s [%(name)28s]: %(message)s [%(threadName)s]"
    fmtStrinfo = "%(asctime)s %(levelname)7s: %(message)s"
    rl = logging.getLogger()
    rl.setLevel(logging.DEBUG)
    
    fileDebugLogH.setFormatter(logging.Formatter(fmtStrDebug))
    fileInfoLogH.setFormatter(logging.Formatter(fmtStrinfo))
    streamLogH.setFormatter(logging.Formatter(fmtStrinfo))
    
    rl.addHandler(fileDebugLogH)
    rl.addHandler(fileInfoLogH)
    rl.addHandler(streamLogH)
     
#     rl.setLevel(logging.DEBUG)
#     rl.addHandler(fileDebugLogH)
#     rl.addHandler(fileInfoLogH)
#     rl.addHandler(streamLogH)

    #logging.basicConfig(level=logging.DEBUG, msgFormat="%(asctime)s %(levelname)s:  %(message)s", handlers=[fileDebugLogH, fileInfoLogH, streamLogH])
    #logging.basicConfig(level=logging.DEBUG, format=fmtStr, handlers=handlers)
        
def parse_known_args_w_help(parser, lastArgs=None):
    '''
    Parse known arguments but show the help message if the parsing fails and the --help or -h was one of the last known arguments 
    :param parser: argparse.ArgumentParser  
    :param lastArgs: the group of arguments known so far 
    '''
    try:
        return parser.parse_known_args()
    except SystemExit:
        if lastArgs != None and lastArgs.help:
            parser.print_help()
            sys.exit(0)
        raise
    


def main():
    defaultResDir = datetime.datetime.now().strftime("%Y_%m_%d_%H_%M_%S")+"_results"

    
    # Creates the parser with the main arguments 
    parser = argparse.ArgumentParser(description=__doc__+"\n\n\n\nAuthor: {}\nVersion: {}".format(__author__,__version__), formatter_class=argparse.RawDescriptionHelpFormatter, add_help=False)
    parser.add_argument("-h", "--help", action='store_true', help='show this help message and exit. Note: if --test is passed this help message will be extended to include test and possibly dependency arguments')
    parser.add_argument("--test", help="the path to a module that implements the test API")
    #parser.add_argument("--test_dependencies", help="the path to the dir that contains the implementation of the test dependencies")
    parser.add_argument("--test_dir", default=defaultResDir, help="path to the test result/working directory. Default: {}".format(defaultResDir))
    setLogArgs(parser)

    args, remaining_argv = parser.parse_known_args() # We parse the known arguments so far
    
    if not args.test:   # --test is a mandatory argument
        if args.help:   # if -h or --help are included in the known arguments so far, print the parser help and exit
            parser.print_help()
            sys.exit(0)
        parser.print_usage()    # else print the usage message and highlight the error, i.e. the argument --test is missing
        print("k3testharness: error: the following arguments are required: --test")
        sys.exit(1)
    
    if not os.path.isfile(args.test):   # if the path of the test file passed does not exist, abort
        logger.error("Given test '{}' with the --test argument does not exist!".format(args.test))
        sys.exit(1)

    #if args.test_dependencies:
    #    if not os.path.isdir(args.test_dependencies):
    #        logger.error("Given test dependency directory '{}' does not exist!".format(args.test_dependencies))
    #        sys.exit(1)
    #    sys.path.append(os.path.abspath(args.test_dependencies))
        
    if not os.path.isdir(args.test_dir) and not args.help:  # if the path to the test result/working directory does not exist, create it 
        os.makedirs(args.test_dir, exist_ok=True)
        
    evalLogArgsAInitLogging(args, args.test_dir) # evaluate the known args and configure the logger
     
    testName = os.path.basename(os.path.splitext(args.test)[0]) # extract the test name from the test filename
    
    try:
        
        # Dynamically imports the module that corresponds to the test
        logger.debug("Instantiating test")
        spec = importlib.util.spec_from_file_location(testName, args.test)
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
    
        # Add a group for test arguments, calling the set_parser_args method defined in the test module
        logger.debug("Get test argparse arguments")
        module.set_parser_args(parser.add_argument_group('test arguments'))
        
        # Parse the known arguments so far (the main arguments and the ones that belong to the test
        args, remaining_argv = parse_known_args_w_help(parser, lastArgs=args)
        
        # Instantiates the test calling the instantiate_test method define in the test module
        testInstance = module.instantiate_test(args.test_dir, args)
        logger.info("Instantiating test successful")
        
    
        # Gets the list of test dependency classes required by the method run_test of the test class in the test module
        testInstanceSignature = inspect.signature(testInstance.run_test)
        testDependencyClasses = [v.annotation for v in testInstanceSignature.parameters.values()]
        logger.info("Test dependencies: "+ ", ".join([c.__name__ for c in testDependencyClasses]))
    
        
        # Adds a group of test dependency arguments for each test dependency class, calling the set_parser_args method defined in each class 
        for dependencyClass in testDependencyClasses:
            if dependencyClass == api.WorkingDirectory:
                continue
            logger.debug("Setting argparse args for dependency {}".format(dependencyClass.__name__))
            dependencyClass.set_parser_args(parser.add_argument_group(dependencyClass.__name__+" arguments"))
        
        # Parse the known arguments so far (the k3testharness arguments, the test arguments and the test dependencies arguments)
        args, remaining_argv = parse_known_args_w_help(parser, lastArgs=args)
        if args.help:
            parser.print_help()
            sys.exit(0)
        
        # If there are still arguments in the remaining_argv group, then they are invalid arguments
        if len(remaining_argv) > 0: #There are unrecognized remaining parameters
            parser.parse_args() # parser_args will exit the program when passing invalid arguments
        
        logger.debug("Parsed argparse arguments '{}'".format(args))
        

        # Gets an instance for all the test dependency classes of the test    
        dependencyInstances = []
        for dependencyClass in testDependencyClasses:
            if dependencyClass == api.WorkingDirectory:
                dependencyInstances.append(str(args.test_dir))
            else:
                logger.debug("Instantiating dependency '{}'".format(dependencyClass.__name__))
                dependencyInstances.append(dependencyClass.instantiate_dependency(args.test_dir, args))
                logger.info("Instantiating dependency '{}' successful".format(dependencyClass.__name__))
    
        logger.info("Instantiating dependencies done")
        
        startDate = datetime.datetime.now()
    
        # For each test dependency, run its corresponding pre_processing methods (sets the dependency state before actually running the test)
        for dependencyInstance in dependencyInstances:
            if hasattr(dependencyInstance, "pre_processing"):
                logger.debug("Dependency {} pre_processing".format(dependencyInstance.__class__.__name__))
                try:
                    dependencyInstance.pre_processing(args.test_dir)
                except:
                    logger.warn("Error while doing pre processing on dependency {}".format(dependencyInstance.__class__.__name__))
                    raise
                logger.info("Dependency {} pre_processing done".format(dependencyInstance.__class__.__name__))
        logger.info("All dependency pre_processing done")
        
        # Performs the actual execution of the test
        try:
            logger.info("Starting Test {} =================================================================".format(testName))
            testInstance.run_test(*dependencyInstances)
        except k3testharness.api.exceptions.TestFailedException as tfe:
            logger.info("Test {}  FAILED =================================================================".format(testName))
            logger.info("EXIT STATUS: NOT OK. TEST FAILED because: {}".format(str(tfe)))
            logger.info("Caught TestFailedException", exc_info=True)
            exitCode = TEST_FAILED
        except Exception as e:
            logger.info("Testing {} FAILED ===============================================================".format(testName))
            logger.info("EXIT STATUS: NOT OK. TESTING FAILED because: {}".format(str(e)))
            logger.info("Caught TestFailedException", exc_info=True)
            exitCode = TESTING_FAILED
        else:
            logger.info("Test {} OK ======================================================================".format(testName))
            logger.info("EXIT STATUS: OK")
            exitCode = TEST_OK
    
        # For each test dependency, run its corresponding post_processing methods
        for dependencyInstance in dependencyInstances:
            if hasattr(dependencyInstance, "post_processing"):
                logger.debug("Dependency {} post_processing".format(dependencyInstance.__class__.__name__))
                try:
                    dependencyInstance.post_processing(args.test_dir)
                except:
                    logger.warn("Error while doing post processing on dependency {}".format(dependencyInstance.__class__.__name__))
                    raise
                logger.info("Dependency {} post_processing done".format(dependencyInstance.__class__.__name__))
        logger.info("All dependency post_processing done")
    
        endDate = datetime.datetime.now()
    
        # Produces a JSON file with test result info
        info = {
            "testName": testName,
            "startDate": str(startDate),
            "endDate": str(endDate),
            "exitStatus": exitCode
        }
        resultFileName = os.path.join(args.test_dir,testName+"_"+datetime.datetime.now().strftime("%Y_%m_%d_%H_%M_%S")+".json")
        with open(resultFileName, 'w') as resultFile:
            resultFile.write(json.dumps(info))
            logger.info("Result file written to: "+resultFileName)
    except Exception as e:
        logger.error("Exception in testharness main. Exiting", exc_info=True)
        exitCode = TESTING_FAILED
    logger.info("Test harness exit code: "+str(exitCode))
    os._exit(exitCode)

if __name__ == '__main__':
    r = main()
    sys.exit(r)
    
    


'''
Created on Mar 25, 2019

@author: Abel Carrión
'''

from abc import abstractmethod, ABCMeta

class Dependency(metaclass=ABCMeta):
    '''
    The Dependency class used as base for implementing test dependencies

    Args:

    Attributes:
    
    '''
    def __init__(self):
        pass
    
    @classmethod
    @abstractmethod
    def set_parser_args(cls, parser):
        '''Sets the arguments that the parser must analyze

        Args:
            parser (argparse.ArgumentParser): the parser instance to be configured

        Raises:

        Returns:
    
        '''
        raise NotImplementedError("Dependency.set_parser_args has no implementation")
    
    @classmethod
    @abstractmethod      
    def instantiate_dependency(cls, testDirectory, args):
        ''''Creates an instance of the Dependency class

        Args:
            args (List of arguments): List of arguments and its corresponding values 
            
        Raises:

        Returns:
            testInstance (Test): A instance of the Test class
    
        '''
        raise NotImplementedError("Dependency.instantiate_dependency has no implementation")
    
    def pre_processing(self, testDir):
        '''Tasks that should be performed BEFORE running the test

        Args: 
            testDir (String): Path to the test result/working directories
            
        Raises:

        Returns:
    
        '''
        pass
    
    def post_processing(self, testDir):
        '''Tasks that should be performed AFTER running the test

        Args: 
            testDir (String): Path to the test result/working directories
            
        Raises:

        Returns:
    
        '''
        pass


'''
Created on 20 Mar 2017

@author: joachim
'''
import datetime, traceback

class Logger:
    
    DEBUG_LVL = 20
    INFO_LVL = 30
    WARN_LVL = 40
    ERROR_LVL = 50
    
    LOG_LVL_TO_STR = {DEBUG_LVL:"debug", INFO_LVL:"info", WARN_LVL:"warning", ERROR_LVL:"error"}
    
    def __init__(self, name):
        self.name = name
    
    def _log(self, logLvl, logItem):
        raise NotImplementedError("getLogger not implemented on base logger")
    
    def debug(self, logMsg):
        self._log(Logger.DEBUG_LVL, logMsg)
    
    def info(self, logMsg):
        self._log(Logger.INFO_LVL, logMsg)
    
    def warning(self, logMsg):
        self._log(Logger.WARN_LVL, logMsg)
    
    def error(self, logMsg):
        self._log(Logger.ERROR_LVL, logMsg)
    
    def getLogger(self, name):
        raise NotImplementedError("getLogger not implemented on base logger")
    
    def close(self):
        pass

class NoneLogger(Logger):
    
    def __init__(self):
        pass
    
    def _log(self, logLvl, logItem):
        pass
    
    def getLogger(self, name):
        return self
    
class LoggerWHandler(Logger):
    
    def __init__(self, name, logEntryHandlers):
        self.name = name
        self.handlers = logEntryHandlers
        
    def _log(self, logLvl, logItem):
        try:
            errLvlStr = str(logLvl) if logLvl not in Logger.LOG_LVL_TO_STR else Logger.LOG_LVL_TO_STR[logLvl]
            errLvlStr = errLvlStr.upper()
            tsStr = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S.%f")[:-3]
            if issubclass(logItem.__class__, Exception):
                for handler in self.handlers:
                    handler.logEntry(logLvl, "{} {:^25.25} {:>7}: Exception Stacktrace:".format(tsStr, self.name, errLvlStr))
                    handler.logEntry(logLvl, "".join(traceback.format_exception(None, logItem, logItem.__traceback__)))
            else:
                for handler in self.handlers:
                    handler.logEntry(logLvl, "{} {:^25.25} {:>7}:  {}".format(tsStr, self.name, errLvlStr, logItem))
        except Exception as e:
            traceback.print_exc()
            print("Exception during logging!")
            print("Log itms was: "+str(logItem))
            
    
    def getLogger(self, name):
        return LoggerWHandler(name, self.handlers)
    
    def close(self):
        for handler in self.handlers:
            if hasattr(handler, "close") and callable(handler.close):
                handler.close()
    
#######################################################
# Handlers
#######################################################
class ConsoleLogHandler:
    def logEntry(self, logLvl, logStr):
        print(logStr)

class StdLogHandler:
    
    def __init__(self, logger):
        self.logger = logger
        
    def logEntry(self, logLvl, logStr):
        getattr(self.logger, Logger.LOG_LVL_TO_STR[logLvl])(logStr)

class MuliLevelLogfileHandler:
    
    def __init__(self, logLvlToFileNmDict):
        self.logLvlToFiles = {}
        for logLvl in logLvlToFileNmDict:
            fh = open(logLvlToFileNmDict[logLvl], "x")
            self.logLvlToFiles[logLvl] = fh
            
    def logEntry(self, logLvl, logStr):
        #print(logStr)
        for fileLogLvl in self.logLvlToFiles:
            if fileLogLvl <= logLvl:
                self.logLvlToFiles[fileLogLvl].write(logStr+"\n")
                self.logLvlToFiles[fileLogLvl].flush()
    
    def close(self):
        for logLvl in self.logLvlToFiles:
            self.logLvlToFiles[logLvl].close()
    
# special loggers
    
            
#######################################################
# Factory Methods
#######################################################

def getConsoleLogger(name):
    return LoggerWHandler(name, [ConsoleLogHandler()])


def getMuliLevelLogfileLoger(name, logLvlToFileNmDict, stdLogger=None):
    return LoggerWHandler(name, [MuliLevelLogfileHandler(logLvlToFileNmDict)] + ([StdLogHandler(stdLogger)] if stdLogger else []))
    
    
    
        


'''
Created on 20 Mar 2017

@author: joachim
'''

class TestFrameworkBaseException(Exception):
    pass

class TestFailedException(TestFrameworkBaseException):
    pass

class TestingFailedException(TestFrameworkBaseException):
    pass

class TestFrameworkException(TestFrameworkBaseException):
    pass

class FrameworkSupportModuleException(TestFrameworkException):
    pass


'''
Created on Mar 22, 2019

@author: Abel Carrión
'''

import sys
from abc import ABCMeta, abstractmethod

def set_parser_args(parser):
    '''Sets the arguments that the parser must analyze

    Args:
        parser (argparse.ArgumentParser): the parser instance to be configured

    Raises:

    Returns:
    
    '''
    raise NotImplemented("Test.setParserArgs has no implementation")
        
def instantiate_test(args):
    '''Creates an instance of the Test class

    Args:
        args (List of arguments): List of arguments and its corresponding values
        testDir (String): Path to the test result/working directories 

    Raises:

    Returns:
        testInstance (Test): A instance of the Test class
    
    '''
    raise NotImplemented("Test.instantiateTest has no implementation")

class Test(metaclass=ABCMeta):
    
    def test_duration_estimate(self):
        '''Gets the estimated test duration in seconds

        Args: 

        Raises:

        Returns:
            duration (integer): the estimated test duration in seconds, None if not defined
    
        '''
        return None
    
    @abstractmethod
    def run_test(self, *dependencies):
        '''Abstract method for implementing the execution of a Test

        Args:
            *dependencies (Objects): A series of dependencies needed for running the Test
            
        Raises:

        Returns:
    
        '''
        raise NotImplemented("Test.run_test has no implementation")


from setuptools import setup, find_packages

setup(
    name='k3testharness',
    version='3.1',
    packages=find_packages(),
    include_package_data=True,
    entry_points={
        'console_scripts': [
            'k3testharness = k3testharness.test_harness:main'
        ]
    }
)